import numpy as np


def newton_type(F, B, z0, tol, maxit):
    """ Implentation of the newtonType method for finding the root of function F,
    i.e. a z for which ||F(z)|| <= tol ≈ 0
    inputs:
    F       vector valued function with identical number of input and
            output arguments, F: R^n -> R^n
            e.g. the gradient of a function
    B       function handle of approximation of jacobian of F
            e.g. the hessian approximation
    z0      initial guess
    tol     tolerance. Values smaller than this are considered 0
    maxit   maximum number of iterations

    outputs:
        Z     iteration history of input variable z
        F_res iteration history of residual of F
    """

    z_dim = np.size(z0)
    Z = np.empty((maxit+2, z_dim))  # array to save history of z
    Z[0] = z0
    F_res = np.empty((maxit+1, z_dim))  # array to save history of residual
    z = z0    # initialize
    for i in range(maxit+1):
        F_eval = np.atleast_1d(F(z))  # evaluate at current iterate
        F_res[i] = F_eval             # save iterate

        # check convergence criterion
        if np.linalg.norm(F_eval, np.inf) <= tol:
            break

        B_eval = np.atleast_2d(B(z))  # if B_eval is not 2D, solve() will raise an error
        z = z - np.linalg.solve(B_eval, F_eval)  # newton type step
        Z[i+1] = z                               # save iterate

    # check max iter
    if i == maxit:
        print("maximum iterations reached")

    # drop the part of array that was not used in iteration.
    Z = Z[0:i+1]
    F_res = F_res[0:i+1]
    return Z, F_res
